//***************************************************************//
//This program is free for use, commercial or non,               //
//you can modify it, redistribute it, anything you want.         //
//It would be nice to credit me (Dennis Kincheloe or DekuTree64) //
//with a mention in the credits or a link to my web site         //
//(http://www.angelfire.com/wizard/deku) if you use it for any   //
//big projects, but it's not required.                           //
//Standard disclaimer: I am not to be held responsible for       //
//anything bad that happens to you or your computer              //
//as a result of using this program                              //
//***************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

typedef unsigned char	u8;
typedef unsigned short	u16;
typedef unsigned long	u32;

typedef signed char		s8;
typedef signed short		s16;
typedef signed long		s32;

typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;

typedef signed char		CHAR;
typedef signed short		SHORT;
typedef signed long		LONG;

#define BIT00 0x1
#define BIT01 0x2
#define BIT02 0x4
#define BIT03 0x8
#define BIT04 0x10
#define BIT05 0x20
#define BIT06 0x40
#define BIT07 0x80
#define BIT08 0x100
#define BIT09 0x200
#define BIT10 0x400
#define BIT11 0x800
#define BIT12 0x1000
#define BIT13 0x2000
#define BIT14 0x4000
#define BIT15 0x8000
#define BIT16 0x10000
#define BIT17 0x20000
#define BIT18 0x40000
#define BIT19 0x80000
#define BIT20 0x100000
#define BIT21 0x200000
#define BIT22 0x400000
#define BIT23 0x800000
#define BIT24 0x1000000
#define BIT25 0x2000000
#define BIT26 0x4000000
#define BIT27 0x8000000
#define BIT28 0x10000000
#define BIT29 0x20000000
#define BIT30 0x40000000
#define BIT31 0x80000000

typedef struct tagRGBQUAD {
        BYTE    rgbBlue;
        BYTE    rgbGreen;
        BYTE    rgbRed;
        BYTE    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER{
        DWORD      biSize;
        LONG       biWidth;
        LONG       biHeight;
        WORD       biPlanes;
        WORD       biBitCount;
        DWORD      biCompression;
        DWORD      biSizeImage;
        LONG       biXPelsPerMeter;
        LONG       biYPelsPerMeter;
        DWORD      biClrUsed;
        DWORD      biClrImportant;
} BITMAPINFOHEADER;

typedef struct tagBITMAPFILEHEADER {
        WORD    bfType;
        DWORD   bfSize;
        WORD    bfReserved1;
        WORD    bfReserved2;
        DWORD   bfOffBits;
} BITMAPFILEHEADER;

typedef struct tOctNode
{
	u32 ref, r, g, b;
	struct tOctNode *child;
} OctNode, *pOctNode;

u32 LoadBmp(char*, BITMAPFILEHEADER*, BITMAPINFOHEADER*, u8**, RGBQUAD**);
u8 **ConvertBmpToTiles(u8*, BITMAPINFOHEADER*, RGBQUAD*);
u8 *AverageTiles(u8**, u32);
u8 *ColorMatch(u8*, u8*, u32, u32);
u8 **MakePalettes(u8**, u32, u8*, u32, u32);
u8 **DitherTiles(u8**, u32, u32, u32, u8**, u8*, u8);
void WriteBmp(char*, u8**, u32, u32, u32, u8**, u32, u32, u8*);
void WriteRaw(char*, u8**, u32, u8**, u32, u32, u8*);
void Quantize(u8*, u32, u8*, u32, u8);

//some globals for the octree functions to use
//so they don't have to pass pointers to eachother all the time
s32 qtMinRef, qtMinMin;
pOctNode qtMinNode;

int Usage()
{
	fprintf(stdout, "Quither v1.2, by Dennis Kincheloe, 2004\n");
	fprintf(stdout, "Useage: quither (-r) pal_cnt pal_size in_file.bmp out_file\n");
	fprintf(stdout, "-r to output raw binary data, otherwise output a bitmap\n");
	fprintf(stdout, "pal_size = 0 for 15 color + 1 transparent (GBA 16-color)\n\n");
	fprintf(stdout, "Example: quither -r 16 0 in.bmp out\n");
	fprintf(stdout, " use 16 palettes, 15 color + 1 transparent each\n");
	fprintf(stdout, " write raw data to out.gfx, out.map, out.pal\n\n");
	fprintf(stdout, "-Notice the output file has no extension,\n");
	fprintf(stdout, " because it depends on the output type\n");
	return 0;
}

/* fgeti() ***************************** 
   Reads a little-endian (Intel-style) integer of len bytes from 
   the given file fp and returns it.  On EOF, sets all high order 
   bits to 0 and sets feof(fp). 
*/ 
unsigned int fgeti(size_t len, FILE *fp) 
{ 
  unsigned int ret = 0, bit_offset = 0; 

  for(; len > 0; len--) 
  { 
    unsigned int c = fgetc(fp); 
    if(c == EOF) 
      return ret; 
    if(bit_offset < 32) 
    { 
      ret |= (c & 0xff) << bit_offset; 
      bit_offset += 8; 
    } 
  } 
  return ret; 
} 


/* fputi() ***************************** 
   Writes a little-endian (Intel-style) integer of len bytes from 
   the given file fp. 
*/ 
void fputi(unsigned long i, size_t len, FILE *fp) 
{ 
  for(; len > 0; len--) 
  { 
    fputc(i & 0xff, fp); 
    i >>= 8; 
  } 
}

int main(int argc, char **argv)
{
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	RGBQUAD *bmColorTable;
	u8 *bmBits;
	s32 i, j, k, palCnt, palSzArg, palSz, realPalSz, 
		tileCnt, bmpW8, bmpH8;
	char szIn[256], szOut[256], outputMode = 'b';
	u8 *img = NULL, **tile = NULL, **tile8 = NULL, *baseColor = NULL, 
		*tilePalIdx = NULL, **pal = NULL;
	FILE *file;

	i = 0;
	for(j = 1; j < argc; j++)
	{
		if(argv[j][0] == '-')
			outputMode = argv[j][1];
		else
		{
			switch(i)
			{
				case 0: palCnt = atoi(argv[j]); break;
				case 1: palSzArg = atoi(argv[j]); break;
				case 2: strcpy(szIn, argv[j]); break;
				case 3: strcpy(szOut, argv[j]); break;
			}
			i++;
		}
	}
	if(i != 4)
		return Usage();

	if(palSzArg == 0)
	{
		palSz = 15;
		realPalSz = 16;
		if(palCnt > 16)
		{
			fprintf(stdout, "\nToo many colors");
			return 0;
		}
	}
	else
	{
		realPalSz = palSz = palSzArg;
		if(palSz * palCnt > 256)
		{
			fprintf(stdout, "\nToo many colors");
			return 0;
		}
	}

	if(LoadBmp(szIn, &bmfh, &bmih, &bmBits, &bmColorTable))
		return 0;

	bmpW8 = bmih.biWidth / 8;
	bmpH8 = bmih.biHeight / 8;

	tileCnt = bmpW8 * bmpH8;

	tile = ConvertBmpToTiles(bmBits, &bmih, bmColorTable);

	delete[] bmBits;
	delete[] bmColorTable;
	img = AverageTiles(tile, tileCnt);
	baseColor = new u8[palCnt*3];
	Quantize(img, tileCnt, baseColor, palCnt, 6);	//quantize the averages
	tilePalIdx = ColorMatch(img, baseColor, tileCnt, palCnt);
	delete[] img;
	pal = MakePalettes(tile, tileCnt, 
		tilePalIdx, palCnt, palSzArg==0?0:palSz);
/*	for(i = 0; i < palCnt; i++)
	{
		for(j = 0; j < realPalSz; j++)
		{
			s32 r, g, b;
			r = ((s32)(pal[i][j*3+0])) - ((s32)(baseColor[i*3+0]));
			r = ((r * 300) >> 8) + ((s32)(baseColor[i*3+0]));
			if(r < 0)
				r = 0;
			else if(r > 63)
				r = 63;
			g = ((s32)(pal[i][j*3+1])) - ((s32)(baseColor[i*3+1]));
			g = ((g * 300) >> 8) + ((s32)(baseColor[i*3+1]));
			if(g < 0)
				g = 0;
			else if(g > 63)
				g = 63;
			b = ((s32)(pal[i][j*3+2])) - ((s32)(baseColor[i*3+2]));
			b = ((b * 300) >> 8) + ((s32)(baseColor[i*3+2]));
			if(b < 0)
				b = 0;
			else if(b > 63)
				b = 63;
			pal[i][j*3+0] = (u8)r;
			pal[i][j*3+1] = (u8)g;
			pal[i][j*3+2] = (u8)b;
		}
	}*/
	tile8 = DitherTiles(tile, tileCnt, bmpW8, bmpH8, 
		pal, tilePalIdx, palSzArg==0?0:realPalSz);

	if(outputMode == 'b')
		WriteBmp(szOut, tile8, tileCnt, bmpW8, bmpH8, 
			pal, palCnt, realPalSz, tilePalIdx);
	else if(outputMode == 'r')
		WriteRaw(szOut, tile8, tileCnt, pal, palCnt, 
			realPalSz, tilePalIdx);

	if(tile8)
	{
		for(i = 0; i < tileCnt; i++)
			if(tile8[i])
				delete[] tile8[i];
		delete[] tile8;
	}
	if(tile)
		for(i = 0; i < tileCnt; i++)
			if(tile[i])
				delete[] tile[i];
	if(pal)
		for(i = 0; i < palCnt; i++)
			if(pal[i])
				delete[] pal[i];
	if(pal)
		delete[] pal;
	if(tile)
		delete[] tile;
	if(tilePalIdx)
		delete[] tilePalIdx;
	if(baseColor)
		delete[] baseColor;

	return 0;
}

u32 LoadBmp(char *szIn, BITMAPFILEHEADER *bmfh, 
	BITMAPINFOHEADER *bmih, u8 **bits, RGBQUAD **cTab)
{
	s32 i;
	FILE *file;
	u8 *bmBits;
	RGBQUAD *bmColorTable;
	file = fopen(szIn, "rb");
	if(!file)
	{
		fprintf(stdout, "\nFile not found");
		return 1;
	}

	bmfh->bfType = fgeti(2, file);
	bmfh->bfSize = fgeti(4, file);
	bmfh->bfReserved1 = fgeti(2, file);
	bmfh->bfReserved2 = fgeti(2, file);
	bmfh->bfOffBits = fgeti(4, file);
	if(bmfh->bfType != 19778)
	{
		fprintf(stdout, "\nNot a bitmap");
		fclose(file);
		return 1;
	}

 	bmih->biSize = fgeti(4, file);
 	bmih->biWidth = fgeti(4, file);
 	bmih->biHeight = fgeti(4, file);
 	bmih->biPlanes = fgeti(2, file);
 	bmih->biBitCount = fgeti(2, file);
 	bmih->biCompression = fgeti(4, file);
 	bmih->biSizeImage = fgeti(4, file);
 	bmih->biXPelsPerMeter = fgeti(4, file);
 	bmih->biYPelsPerMeter = fgeti(4, file);
 	bmih->biClrUsed = fgeti(4, file);
 	bmih->biClrImportant = fgeti(4, file);
	if((bmih->biBitCount != 24 && bmih->biBitCount != 8))
	{
		fprintf(stdout, "\nNot an 8 or 24-bit BMP");
		fclose(file);
		return 1;
	}
	if((bmih->biWidth & 7) || (bmih->biHeight & 7))
		fprintf(stdout, "\nDimensions cropped to multiples of 8");

	if(bmih->biBitCount == 8)
	{
		bmColorTable = new RGBQUAD[256];
		for(i = 0; i < 256; i++)
		{
			bmColorTable[i].rgbBlue = fgeti(1, file);
			bmColorTable[i].rgbGreen = fgeti(1, file);
			bmColorTable[i].rgbRed = fgeti(1, file);
			bmColorTable[i].rgbReserved = fgeti(1, file);
		}
		bmBits = new u8[bmih->biWidth * bmih->biHeight];
		if(!bmBits)
		{
			fprintf(stdout, "\nOut of memory");
			fclose(file);
			return 1;
		}
		for(i = 0; i < bmih->biWidth * bmih->biHeight; i++)
			bmBits[i] = fgeti(1, file);
	}
	else if(bmih->biBitCount == 24)
	{
		bmColorTable = NULL;
		bmBits = new u8[bmih->biWidth * bmih->biHeight * 3];
		if(!bmBits)
		{
			fprintf(stdout, "\nOut of memory");
			fclose(file);
			return 1;
		}
		for(i = 0; i < bmih->biWidth * bmih->biHeight * 3; i++)
			bmBits[i] = fgeti(1, file);
	}
	fclose(file);
	*bits = bmBits;
	*cTab = bmColorTable;
	return 0;
}

u8 **ConvertBmpToTiles(u8 *bmBits, 
	BITMAPINFOHEADER *bmih, RGBQUAD *bmColorTable)
{
	s32 i, j, k, bmpW8, bmpH8, bmpWidthBytes, tileCnt, yOfs;
	u8 **tile, *img;

	bmpW8 = bmih->biWidth / 8;
	bmpH8 = bmih->biHeight / 8;

	tileCnt = bmpW8 * bmpH8;
	tile = new u8*[tileCnt];
	memset(tile, 0, tileCnt*4);

	if(bmih->biBitCount == 8)
	{
		bmpWidthBytes = (bmih->biWidth + 3) & ~3;
		yOfs = bmih->biHeight - bmpH8*8;
		for(i = 0; i < tileCnt; i++)
		{
			tile[i] = new u8[64*3];
			img = (u8*)bmBits;
			img += (i % bmpW8) * 8 + 
				((tileCnt - 1 - i) / bmpW8 * 8 + yOfs) * bmpWidthBytes;
			for(j = 0; j < 8; j++)
			{
				for(k = 0; k < 8; k++)
				{
					tile[i][(7-j)*8*3+k*3] = bmColorTable[*img].rgbBlue >> 2;
					tile[i][(7-j)*8*3+k*3+1] = bmColorTable[*img].rgbGreen >> 2;
					tile[i][(7-j)*8*3+k*3+2] = bmColorTable[*img].rgbRed >> 2;
					img++;
				}
				img += (bmpWidthBytes - 8);
			}
		}
	}
	else if(bmih->biBitCount == 24)
	{
		bmpWidthBytes = (bmih->biWidth*3 + 3) & ~3;
		yOfs = bmih->biHeight - bmpH8*8;
		for(i = 0; i < tileCnt; i++)
		{
			tile[i] = new u8[64*3];
			img = (u8*)bmBits;
			img += (i % bmpW8) * 8*3 + 
				((tileCnt - 1 - i) / bmpW8 * 8 + yOfs) * bmpWidthBytes;
			for(j = 0; j < 8; j++)
			{
				for(k = 0; k < 8*3; k++)
				{
					tile[i][(7-j)*8*3+k] = *img++ >> 2;	//convert to 6-bit (seems to be the best speed/quality compromise)
				}
				img += bmpWidthBytes - 8*3;
			}
		}
	}
	return tile;
}

u8 *AverageTiles(u8 **tile, u32 tileCnt)
{
	s32 i, j;
	u32 *tmp = new u32[tileCnt * 3];
	u8 *img;
	for(i = 0; i < tileCnt; i++)
	{
		tmp[i * 3] = tmp[i * 3 + 1] = tmp[i * 3 + 2] = 0;
		for(j = 0; j < 64; j++)
		{
			tmp[i * 3] += tile[i][j * 3];
			tmp[i * 3 + 1] += tile[i][j * 3 + 1];
			tmp[i * 3 + 2] += tile[i][j * 3 + 2];
		}
	}
	img = new u8[tileCnt*3];
	for(i = 0; i < tileCnt * 3; i++)
		img[i] = tmp[i] / 64;	//average all the colors from this tile
	delete[] tmp;
	return img;
}

u8 *ColorMatch(u8 *img, u8 *baseColor, u32 tileCnt, u32 palCnt)
{
	s32 i, j;
	u8 *tilePalIdx = new u8[tileCnt];
	//and find the closest color to this tile's original average
	for(i = 0; i < tileCnt; i++)
	{
		s32 dist = 256*256*3;
		tilePalIdx[i] = 0xff;
		for(j = 0; j < palCnt; j++)
		{
			s32 r, g, b, c;
			r = baseColor[j*3];
			r -= img[i*3];
			g = baseColor[j*3+1];
			g -= img[i*3+1];
			b = baseColor[j*3+2];
			b -= img[i*3+2];
			c = r*r + g*g + b*b;
			if(c < dist)
			{
				dist = c;
				tilePalIdx[i] = j;
			}
		}
	}
	return tilePalIdx;
}

u8 **MakePalettes(u8 **tile, u32 tileCnt, 
	u8 *tilePalIdx, u32 palCnt, u32 palSz)
{
	s32 i, j, k;
	u8 *img, **pal;
	//copy all the tiles that use each average, and then quantize
	pal = new u8*[palCnt];
	memset(pal, 0, palCnt*4);
	for(i = 0; i < palCnt; i++)
	{
		k = 0;
		for(j = 0; j < tileCnt; j++)
			if(tilePalIdx[j] == i)
				k++;
		img = new u8[k*64*3];
		k = 0;
		for(j = 0; j < tileCnt; j++)
		{
			if(tilePalIdx[j] == i)
			{
				memcpy(img + k*3, tile[j], 64*3);
				k += 64;
			}
		}
		pal[i] = new u8[(palSz==0?16:palSz)*3];
		if(palSz == 0)
		{
			pal[i][0] = 63;
			pal[i][1] = 0;
			pal[i][2] = 63;
			Quantize(img, k, pal[i] + 3, 15, 6);
		}
		else
			Quantize(img, k, pal[i], palSz, 6);
		delete[] img;
	}
	return pal;
}

u8 **DitherTiles(u8 **tile, u32 tileCnt, u32 bmpW8, u32 bmpH8, 
	u8 **pal, u8 *tilePalIdx, u8 palSz)
{
	s32 i, j, k, k2;
	u8 **tile8;
	s8 **tErr;

	tErr = new s8*[tileCnt];
	tile8 = new u8*[tileCnt];
	for(i = 0; i < tileCnt; i++)
	{
		tile8[i] = new u8[64];
		tErr[i] = new s8[64*3];
		memset(tile8[i], 0, 64);
		memset(tErr[i], 0, 64*3);
	}
	for(i = 0; i < bmpH8*8; i++)
	{
		for(j = 0; j < bmpW8*8; j++)
		{
			s32 ii, jj, ip1, jp1;
			s32 t = (i>>3)*bmpW8 + (j>>3), px = ((i&7)<<3) + (j&7);
			u8 *curPal = pal[tilePalIdx[t]];
			s32 r, g, b, re = 0, ge = 0, be = 0, c = 0, dist = 1<<30;
			r = tile[t][px*3];
			r += tErr[t][px*3] >> 4;
			g = tile[t][px*3+1];
			g += tErr[t][px*3+1] >> 4;
			b = tile[t][px*3+2];
			b += tErr[t][px*3+2] >> 4;
			for(k = palSz==0?1:0, k2 = palSz==0?16:palSz; k < k2; k++)
			{
				s32 r2, g2, b2, rmean, d;
				r2 = curPal[k*3];
				rmean = (r + r2) >> 1;
				r2 = r - r2;
				g2 = curPal[k*3+1];
				g2 = g - g2;
				b2 = curPal[k*3+2];
				b2 = b - b2;
				//d = r2*r2 + g2*g2 + b2*b2;
				d = (((512+rmean)*r2*r2)>>8) + 4*g2*g2 + 
					(((768-rmean)*b2*b2)>>8);	//better color distance function
				if(d < dist)
				{
					re = r2;
					ge = g2;
					be = b2;
					dist = d;
					c = k;
				}
			}
			tile8[t][px] = c;// + tilePalIdx[t]*(palSz==0?16:palSz);
			if(j+1 < bmpW8*8)
			{
				t = (i>>3)*bmpW8 + ((j+1)>>3);
				px = ((i&7)<<3) + ((j+1)&7);
				tErr[t][px*3] += re*7;
				tErr[t][px*3+1] += ge*7;
				tErr[t][px*3+2] += be*7;
			}
			if(i+1 < bmpH8*8)
			{
				if(j > 0)
				{
					t = ((i+1)>>3)*bmpW8 + ((j-1)>>3);
					px = (((i+1)&7)<<3) + ((j-1)&7);
					tErr[t][px*3] += re*3;
					tErr[t][px*3+1] += ge*3;
					tErr[t][px*3+2] += be*3;
				}
				t = ((i+1)>>3)*bmpW8 + (j>>3);
				px = (((i+1)&7)<<3) + (j&7);
				tErr[t][px*3] += re*5;
				tErr[t][px*3+1] += ge*5;
				tErr[t][px*3+2] += be*5;
				if(j+1 < bmpW8*8)
				{
					t = ((i+1)>>3)*bmpW8 + ((j+1)>>3);
					px = (((i+1)&7)<<3) + ((j+1)&7);
					tErr[t][px*3] += re;
					tErr[t][px*3+1] += ge;
					tErr[t][px*3+2] += be;
				}
			}
		}
	}
	for(i = 0; i < tileCnt; i++)
		delete[] tErr[i];
	delete[] tErr;
//non-dithered color matching
/*	{
		for(i = 0; i < tileCnt; i++)
		{
			u8 *curPal = pal[tilePalIdx[i]];
			tile8[i] = new u8[64];
			for(j = 0; j < 64; j++)
			{
				s32 r, g, b, c = 0, dist = 32*32*3;
				r = tile[i][j*3];
				g = tile[i][j*3+1];
				b = tile[i][j*3+2];
				for(k = 0; k < palSz; k++)
				{
					s32 r2, g2, b2, d;
					r2 = curPal[k*3];
					r2 -= r;
					g2 = curPal[k*3+1];
					g2 -= g;
					b2 = curPal[k*3+2];
					b2 -= b;
					d = r2*r2 + g2*g2 + b2*b2;
					if(d < dist)
					{
						dist = d;
						c = k;
					}
				}
				tile8[i][j] = c + tilePalIdx[i] * realPalSz;
			}
		}
	}*/
	return tile8;
}

void WriteBmp(char *szOut, u8 **tile8, u32 tileCnt, u32 bmpW8, u32 bmpH8, 
	u8 **pal, u32 palCnt, u32 palSz, u8 *tilePalIdx)
{
	s32 i, j, k;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	u32 bmpPal[256];
	FILE *file;
	char szTemp[256];

	strcpy(szTemp, szOut);
	strcat(szTemp, ".bmp");

	file = fopen(szTemp, "wb");
	bmfh.bfType = 0x4d42;
	bmfh.bfSize = 1078 + tileCnt * 64;
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = 1078;
	fputi(bmfh.bfType, 2, file);
	fputi(bmfh.bfSize, 4, file);
	fputi(bmfh.bfReserved1, 2, file);
	fputi(bmfh.bfReserved2, 2, file);
	fputi(bmfh.bfOffBits, 4, file);
	bmih.biSize = 40;
	bmih.biWidth = bmpW8*8;
	bmih.biHeight = bmpH8*8;
	bmih.biPlanes = 1;
	bmih.biBitCount = 8;
	bmih.biCompression = 0;
	bmih.biSizeImage = tileCnt * 64;
	bmih.biXPelsPerMeter = 0;
	bmih.biYPelsPerMeter = 0;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	fputi(bmih.biSize, 4, file);
	fputi(bmih.biWidth, 4, file);
	fputi(bmih.biHeight, 4, file);
	fputi(bmih.biPlanes, 2, file);
	fputi(bmih.biBitCount, 2, file);
	fputi(bmih.biCompression, 4, file);
	fputi(bmih.biSizeImage, 4, file);
	fputi(bmih.biXPelsPerMeter, 4, file);
	fputi(bmih.biYPelsPerMeter, 4, file);
	fputi(bmih.biClrUsed, 4, file);
	fputi(bmih.biClrImportant, 4, file);

	memset(bmpPal, 0, 1024);
	for(i = 0; i < palCnt; i++)
	{
		for(j = 0; j < palSz; j++)
		{
			bmpPal[i*palSz + j] = 
				((pal[i][j*3] << 2) | (pal[i][j*3] >> 4)) | 
				((pal[i][j*3+1] << 10) | (pal[i][j*3+1] >> 4 << 8)) | //shift 6-bit colors
				((pal[i][j*3+2] << 18) | (pal[i][j*3+2] >> 4 << 16));	//back up to 8-bit
		}
	}
	for(i = 0; i < 256; i++)
		fputi(bmpPal[i], 4, file);

	for(i = bmpH8 - 1; i >= 0; i--)
	{
		for(j = 7; j >= 0; j--)
		{
			for(k = 0; k < bmpW8; k++)
			{
				u32 palOfs = palSz * tilePalIdx[i * bmpW8 + k];
				fputi(tile8[i * bmpW8 + k][j*8+0] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+1] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+2] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+3] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+4] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+5] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+6] + palOfs, 1, file);
				fputi(tile8[i * bmpW8 + k][j*8+7] + palOfs, 1, file);
			}
		}
	}
	fclose(file);
}

void WriteRaw(char *szOut, u8 **tile8, u32 tileCnt, 
	u8 **pal, u32 palCnt, u32 palSz, u8 *tilePalIdx)
{
	s32 i, j, k;
	FILE *file;
	char szTemp[256];

	strcpy(szTemp, szOut);
	strcat(szTemp, ".gfx");
	file = fopen(szTemp, "wb");

	if(palSz <= 16)
		for(i = 0; i < tileCnt; i++)
			for(j = 0; j < 32; j++)
				fputi(tile8[i][j << 1] | 
					(tile8[i][(j << 1) + 1] << 4), 1, file);
	else
		for(i = 0; i < tileCnt; i++)
			for(j = 0; j < 64; j++)
				fputi(tile8[i][j], 1, file);
	fclose(file);

	strcpy(szTemp, szOut);
	strcat(szTemp, ".map");
	file = fopen(szTemp, "wb");
	if(palSz <= 16)
		for(i = 0; i < tileCnt; i++)
			fputi(i | (tilePalIdx[i] << 12), 2, file);
	else
		for(i = 0; i < tileCnt; i++)
			fputi(i, 2, file);
	fclose(file);

	strcpy(szTemp, szOut);
	strcat(szTemp, ".pal");
	file = fopen(szTemp, "wb");
	for(i = 0; i < palCnt; i++)
		for(j = 0; j < palSz; j++)
			fputi((pal[i][j*3] >> 1 << 10) | (pal[i][j*3+1] >> 1 << 5) | 
				(pal[i][j*3+2] >> 1), 2, file);
	fclose(file);
}

void OctreeInit(pOctNode node)
{
	s32 i;
	node->child = new OctNode[8];
	for(i = 0; i < 8; i++)
	{
		node->child[i].ref = node->child[i].r = 
			node->child[i].g = node->child[i].b = 0;
		node->child[i].child = NULL;
	}
}

void OctreeDelete(pOctNode node)
{
	s32 i;
	if(node->child)
	{
		for(i = 0; i < 8; i++)
			OctreeDelete(&node->child[i]);
		delete[] node->child;
		node->child = NULL;
	}
}

void OctreeIncColor(pOctNode node, u32 color, s8 bit, u32 *leafCnt)
{
	if(bit >= 0)
	{
		if(!node->child)
			OctreeInit(node);
		u32 idx = color >> bit;
		idx = (idx & 1) | ((idx >> 7) & 2) | ((idx >> 14) & 4);
		OctreeIncColor(&node->child[idx], color, bit - 1, leafCnt);
	}
	else
	{
		if(node->ref == 0)
			*leafCnt += 1;
		node->ref++;
		node->r += color & 0xff;
		node->g += (color >> 8) & 0xff;
		node->b += color >> 16;
	}
}

//this function is where most of the time is spent
//I tried to speed it up as best I could, but it's still pretty slow
u32 OctreeGetMinNode(pOctNode node)
{
	s32 i, ref = 0;
	if(node->child)
	{
		for(i = 7; i >= 0; i--)
		{
			s32 tmp = OctreeGetMinNode(&node->child[i]);
			if(tmp == 0xffffffff)
				return 0xffffffff;
			ref += tmp;
		}
		if(ref < qtMinRef)
		{
			qtMinRef = ref;
			qtMinNode = node;
			if(ref <= qtMinMin)
				return 0xffffffff;
		}
		return ref;
	}
	else
		return node->ref;
}

void OctreeRemoveNode(pOctNode node, u32 *leafCnt, u32 palSz)
{
	s32 i;
	if(node->child)
	{
		*leafCnt += 1;
		for(i = 7; i >= 0; i--)
		{
			if(node->child[i].ref > 0)
			{
				*leafCnt -= 1;
				node->ref += node->child[i].ref;
				node->r += node->child[i].r;
				node->g += node->child[i].g;
				node->b += node->child[i].b;
				if(*leafCnt <= palSz)
				{
					if(i > 0)
					{
						for(s32 j = 7; j >= i; j--)
						{
							node->child[j].ref = 0;
							node->child[j].r = 0;
							node->child[j].g = 0;
							node->child[j].b = 0;
						}
						return;
					}
				}
			}
		}
		delete[] node->child;
		node->child = NULL;
	}
}

void OctreeMakePal(pOctNode node, u8 *pal, u32 palSz, u32 *curSz)
{
	s32 i;
	if(node->child)
	{
		for(i = 0; i < 8; i++)
			OctreeMakePal(&node->child[i], pal, palSz, curSz);
	}
	if(node->ref > 0 && *curSz < palSz)
	{
		pal[*curSz * 3] = node->r / node->ref;
		pal[*curSz * 3 + 1] = node->g / node->ref;
		pal[*curSz * 3 + 2] = node->b / node->ref;
		*curSz += 1;
	}
}

void Quantize(u8 *img, u32 imgSz, u8 *pal, u32 palSz, u8 bit)
{
	s32 i;
	u32 leafCnt = 0;
	OctNode tree;

	tree.ref = tree.r = tree.g = tree.b = 0;
	tree.child = NULL;

	for(i = 0; i < imgSz; i++)
		OctreeIncColor(&tree, img[i * 3] | 
			(img[i * 3 + 1] << 8) | (img[i * 3 + 2] << 16), bit - 1, &leafCnt);
	qtMinMin = 1;
	while(leafCnt > palSz)
	{
		qtMinRef = imgSz + 1;
		OctreeGetMinNode(&tree);
		qtMinMin = qtMinRef;
		OctreeRemoveNode(qtMinNode, &leafCnt, palSz);
	}
	i = 0;
	OctreeMakePal(&tree, pal, palSz, (u32*)(&i));
	OctreeDelete(&tree);
}
